from setuptools import setup, find_packages
from pathlib import Path
import glob

PACKAGE_NAME = 'pygidsim'

if __name__ == '__main__':
    setup(
        name=PACKAGE_NAME,
        version='0.1.1',
        description='Simulation of wide-angle grazing-incidence diffraction patterns',
        long_description=(Path(__file__).parent / "README.md").read_text(encoding="utf-8"),
        long_description_content_type="text/markdown",
        author='Mikhail Romodin',
        url='https://github.com/mlgid-project/pygidSIM',
        packages=find_packages(),
        python_requires='>=3.8',
        install_requires=[
            'numpy>=1.24.4',
            'xrayutilities==1.7.10',
            'scipy>=1.10.1',
            'psutil>=5.9.0'],
        extras_require={
            'dev': [
                'pytest>=7.0.0',
                'pytest-cov>=4.0.0',
                'pytest-xdist>=3.0.0',
                'pytest-mock>=3.10.0',
                'black>=23.0.0',
                'flake8>=6.0.0',
                'isort>=5.12.0',
                'mypy>=1.0.0',
                'pre-commit>=3.0.0',
            ],
            'test': [
                'pytest>=7.0.0',
                'pytest-cov>=4.0.0',
                'pytest-xdist>=3.0.0',
                'pytest-mock>=3.10.0',
            ]
        },
        data_files=glob.glob('pygidsim/data/**'),
    )
